/*
 * (c) 2006-2007 Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

/* To compile, we need SUS/XPG4v2 or later. */
/*#define _XOPEN_SOURCE 500*/

#include <alloca.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <math.h>
#include <signal.h>
#ifdef HAVESPAWN
#include <spawn.h>
#endif
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

/* SUSv3 does not have PATH_MAX */
#ifndef PATH_MAX
#define PATH_MAX _XOPEN_PATH_MAX
#endif

#if __STDC_VERSION__ < 199901L
#define inline
#endif

#include "log.h"
#include "tokens.h"


typedef struct arg_s {
	char *arg;
	struct arg_s *prev;
} arg_t;

typedef struct job_s {
	pid_t pid;
	char **args;
	size_t numargs;
	char *in, *out, *err, *cmd;
	int outfile;
	int append;	/* 1 = stdout, 2 = stderr, 3 = both */
	unsigned long long start;
	struct job_s *next;
} job_t;

static arg_t *Baseargs = 0;
static size_t Numbaseargs = 0, Running = 0, Waiting = 0, Pagesize;
static long Limit = 0;
job_t *Executing = 0;
static job_t *Ready = 0, *ReadyEnd = 0, *JobCache = 0;
static arg_t *ArgCache = 0;
static int Stdout = -1, Stderr = -1, Stdin = -1, Prompt = 0;
static FILE *Tty = 0;
static char *Path;

extern int yylex(void);
extern verbose_t Verbose;


static void *Malloc(size_t s)
{
	int tries = 0, time = 5;
	for (;;) {
		void *r = malloc(s);
		if (r != 0)
			return r;
		warn("out of memory - retrying in %d seconds\n",time);
		(void) sleep(time);
		time += 5;
		++tries;
		if (tries == 10) {
			warn("Persisiting low memory situation: stopping.\n"
				"Please send SIGCONT to contiue.\n");
			if (-1 == raise(SIGSTOP))
				abort();
			time = 5;
			tries = 0;
		}
	}
}


#ifdef __sun
#define gettime() (gethrtime()/1000)
#else
static inline unsigned long long gettime(void)
{
	struct timeval now;
	
	(void) gettimeofday(&now,0);
	return now.tv_sec * 1000000 + now.tv_usec;
}
#endif


static inline void close_onexec(int f)
{
	if (f != -1) {
		int err = fcntl(f,F_SETFD,FD_CLOEXEC);
		assert(err != -1);
	}
}


static inline char *Strdupn(const char *s, size_t n)
{
	assert(n);
	return memcpy(Malloc(n),s,n);
}


static void name_job(job_t *j)
{
	char **a = j->args, *c;
	size_t l = 0;
	if (j->cmd)
		return;
	assert(a);
	do {
		l += strlen(*a);
		++l;
	} while (*++a);
	j->cmd = c = Malloc(l);
	a = j->args;
	do {
		c = (char *) memccpy(c,*a,0,l);
		*--c = ' ';
		++c;
	} while (*++a);
	--c;
	*c = 0;
}


static void exec_job(job_t *j)
{
	pid_t pid;
	int in = -1, out = -1, err = -1, log = -1, ret;
#ifdef HAVESPAWN
	posix_spawn_file_actions_t acts;
#endif

	if (Prompt) {
		char y[16];
		name_job(j);
		(void) fprintf(Tty,"start job %s?\n",j->cmd);
		(void) fgets(y,sizeof(y),Tty);
		if (ferror(Tty))
			error("cannot read from terminal\n");
		if (strcasecmp(y,"yes\n") && strcasecmp(y,"y\n")) {
			j->pid = 0;
			return;
		}
	}

	j->start = gettime();
#ifndef HAVESPAWN
	if (j->in) {	/* stdin is redirected */
		in = open(j->in,O_RDONLY,0666);
		if (in == -1) {
			warn("cannot open input file %s: %s\n",j->in,strerror(errno));
			return;
		}
		dbug("job input redirected to %s\n",j->in);
	} else {
		in = Stdin;
	}
	if (j->err) {	/* stdout and stderr are redirected */
		int mode;
	       	if (j->append & 2)
			mode = O_WRONLY|O_CREAT|O_APPEND;
		else
			mode = O_WRONLY|O_CREAT|O_TRUNC|O_EXCL;
		err = open(j->err,mode,0666);
		if (err == -1) {
			warn("cannot open error output file %s: %s\n",j->err,strerror(errno));
			return;
		}
		out = err;
		dbug("job error output redirected to %s\n",j->err);
	} else if (Stdout != -1) {
		err = Stderr;
	}
	if (j->out) {	/* output is redirected */
		int mode;
	       	if (j->append & 1)
			mode = O_WRONLY|O_CREAT|O_APPEND;
		else
			mode = O_WRONLY|O_CREAT|O_TRUNC|O_EXCL;
		out = open(j->out,mode,0666);
		if (out == -1) {
			warn("cannot open output file %s: %s\n",j->out,strerror(errno));
			return;
		}
		dbug("job output redirected to %s\n",j->out);
	} else if (Stdout != -1) {
		out = Stdout;
	}
	if (err == -1) {
		char fname[] = "/tmp/xjobs-XXXXXX";
		err = mkstemp(fname);
		if (-1 == err)
			error("cannot create intermediate output file %s: %s\n",fname,strerror(errno));
		j->outfile = open(fname,O_RDONLY);
		if (-1 == j->outfile)
			error("cannot access intermediate output file: %s\n",strerror(errno));
		close_onexec(j->outfile);
		(void) unlink(fname);
		if (out == -1)
			out = err;
	}
	pid = vfork();
	if (pid == -1)
		error("error could not fork: %s\n",strerror(errno));
	if (pid != 0) {
		j->pid = pid;
		name_job(j);
		if (Verbose >= Info)
			(void) printf("### started pid %ld: %s\n",pid,j->cmd);
		j->next = Executing;
		if ((in != -1) && (in != Stdin))
			(void) close(in);
		if ((out != -1) && (out != Stdout))
			(void) close(out);
		if ((err != -1) && (err != Stderr))
			(void) close(err);
		Executing = j;
		if (--Waiting && Verbose)
			dbug("%d jobs still waiting in the queue\n",Waiting);
		++Running;
	} else {
		int r;
		if (j->outfile != -1)
			(void) close(j->outfile);
		if (in != -1) {
			if (-1 == dup2(in,STDIN_FILENO))
				warn("unable to redirect stdin: %s\n",strerror(errno));
			if (j->in == 0) {
				int e = fcntl(STDIN_FILENO,F_SETFD,0);
				assert(e != -1);
			}
		}
		if (out != -1) {
			if (-1 == dup2(out,STDOUT_FILENO))
				warn("unable to redirect stdout: %s\n",strerror(errno));
			if (j->out == 0) {
				int e = fcntl(STDOUT_FILENO,F_SETFD,0);
				assert(e != -1);
			}
		}
		if (err != -1) {
			if (-1 == dup2(err,STDERR_FILENO))
				warn("unable to redirect stderr: %s\n",strerror(errno));
			if (j->err == 0) {
				int e = fcntl(STDERR_FILENO,F_SETFD,0);
				assert(e != -1);
			}
		}
		r = execv(j->args[0],j->args);
		assert(r == -1);
		error("failed to execute %s: %s\n",j->args[0],strerror(errno));
		exit(EXIT_FAILURE);
	}
#else
	ret = posix_spawn_file_actions_init(&acts);
	assert(ret == 0);
	if (j->in) {
		in = open(j->in,O_RDONLY);
		if (in == -1)
			warn("cannot open input file %s: %s\n",j->in,strerror(errno));
		ret = posix_spawn_file_actions_addopen(&acts,STDIN_FILENO,j->in,O_RDONLY,0666);
	} else if (Stdin != -1) {
		ret = fcntl(Stdin,F_SETFD,0);
		assert(ret != -1);
		in = Stdin;
	}
	if (in != -1) {
		ret = posix_spawn_file_actions_adddup2(&acts,in,STDIN_FILENO);
		assert(ret == 0);
	}
	if (j->err) {
		int mode;
		if (j->append & 2)
			mode = O_WRONLY|O_APPEND|O_CREAT;
		else
			mode = O_WRONLY|O_CREAT|O_TRUNC|O_EXCL;
		err = open(j->err,mode,0666);
		if (err == -1) {
			warn("cannot open error output file %s: %s\n",j->err,strerror(errno));
			return;
		}
	} else if (Stderr != -1) {
		ret = fcntl(Stderr,F_SETFD,0);
		assert(ret != -1);
		err = Stderr;
	} else {
		char fname[] = "/tmp/xjobs-XXXXXX";
		log = mkstemp(fname);
		if (-1 == log)
			error("cannot create intermediate errorlog file %s: %s\n",fname,strerror(errno));
		j->outfile = open(fname,O_RDONLY);
		if (-1 == j->outfile)
			error("cannot access intermediate errorlog file: %s\n",strerror(errno));
		close_onexec(j->outfile);
		(void) unlink(fname);
		err = log;
	}
	ret = posix_spawn_file_actions_adddup2(&acts,err,STDERR_FILENO);
	assert(ret == 0);
	if (j->out) {
		int mode;
		if (j->append & 1)
			mode = O_WRONLY|O_APPEND|O_CREAT;
		else
			mode = O_WRONLY|O_CREAT|O_TRUNC|O_EXCL;
		out = open(j->err,mode,0666);
		if (out != -1) {
			warn("cannot open output file %s: %s\n",j->out,strerror(errno));
			return;
		}
	} else if (Stdout != -1) {
		ret = fcntl(Stdout,F_SETFD,0);
		assert(ret != -1);
		out = Stdout;
	} else {
		assert(log != -1);
		out = log;
	}
	ret = posix_spawn_file_actions_adddup2(&acts,out,STDOUT_FILENO);
	assert(ret == 0);

	if (posix_spawn(&j->pid,j->args[0],&acts,0,j->args,0) == 0) {
		name_job(j);
		if (Verbose >= Info)
			(void) printf("### started pid %ld: %s\n",j->pid,j->cmd);
		j->next = Executing;
		if ((in != -1) && (in != Stdin))
			(void) close(in);
		if ((out != -1) && (out != Stdout))
			(void) close(out);
		if ((err != -1) && (err != Stderr) && (err != out))
			(void) close(err);
		Executing = j;
		if (--Waiting && Verbose)
			dbug("%d jobs still waiting in the queue\n",Waiting);
		++Running;
	} else {
		warn("couldn't spawn job %s: %s\n",j->cmd,strerror(errno));
		j->pid = 0;
	}
	(void) posix_spawn_file_actions_destroy(&acts);
#endif
}


static int resolve_symlink(char *cmd)
{
	char lt[PATH_MAX];
	ssize_t l;
	/* handle symlinks */
	bzero(lt,sizeof(lt));
	while (-1 != (l = readlink(cmd,lt,PATH_MAX-1))) {
		lt[l] = '\0';
		dbug("is symlink: %s\n",lt);
		/* handle relative links */
		if (lt[0] != '/') {
			char *base = strrchr(cmd,'/');   
			if (base == 0) {
				base = cmd;
			} else if ((strlen(lt) + (base - cmd)) >= PATH_MAX) {
				warn("command exceed PATH_MAX limit of your system\n");
				return -1;
			}
			++base;
			(void) strcpy(base,lt);
			dbug("resolved to: %s\n",cmd);
		} else
			(void) strcpy(cmd,lt);
	}
	return 0;
}


static char *complete_exe(char *exe)
{
	typedef struct cache_s {
		char *exe, *file;
		struct cache_s *next;
	} cache_t;
	static cache_t *cache = 0;
	cache_t *c = cache;
	char *token, *next;
	size_t exel;

	while (c) {
		if (0 == strcmp(exe,c->exe)) {
			if (c->file == c->exe)
				return exe;
			return strdup(c->file);
		}
		c = c->next;
	}
	if ((exe[0] == '/') && (0 == access(exe,X_OK))) {
		c = malloc(sizeof(cache_t));
		c->exe = (char *) strdup(exe);
		c->file = c->exe;
		c->next = cache;
		cache = c;
                return exe;
	}
	exel = strlen(exe);
	if ((strncmp(exe,"./",2) == 0) || (strncmp(exe,"../",3) == 0)) {
		if (0 == access(exe,X_OK)) {
			c = malloc(sizeof(cache_t));
			c->exe = (char *) strdup(exe);
			c->file = c->exe;
			c->next = cache;
			cache = c;
			return exe;
		}
		if (Verbose == Debug) {
			char cmd[PATH_MAX];
			(void) memcpy(cmd,exe,exel);
			(void) resolve_symlink(cmd);
			return 0;
		}
	}
	token = Path;
	next = strchr(Path,':');
        while (token && *token) {
		char cmd[PATH_MAX];
		size_t off;

		if (next)
			*next = 0;
		if ((off = strlen(token)) >= PATH_MAX) {
			warn("command exceed PATH_MAX limit of your system\n");
			return 0;
		}
		(void) strcpy(cmd, token);
		if (cmd[off-1] != '/') {
			cmd[off] = '/';
			++off;
		}
		if (exel + off >= PATH_MAX) {
			warn("command exceed PATH_MAX limit of your system\n");
			return 0;
		}
		(void) strcpy(cmd+off, exe);
		if (next) {
			*next = ':';
			token = next + 1;
			next = strchr(token,':');
		} else
			token = 0;
		dbug("candidate: %s\n",cmd);
		if ((Verbose == Debug) && resolve_symlink(cmd))
			continue;
		if (0 == access(cmd,X_OK)) {
			dbug("using: %s\n",cmd);
			c = malloc(sizeof(cache_t));
			c->exe = (char *) strdup(exe);
			c->file = (char *) strdup(cmd);
			c->next = cache;
			cache = c;
                        return strdup(cmd);
		}
        }
        return 0;
}


static void start_jobs(void)
{
	while ((Running < Limit) && Ready) {
		job_t *j = Ready;
		if ((j->args == 0) && (Running > 0)) {
			/* sequence point */
			dbug("sequence point - waiting for all processes to finish\n");
			return;
		}
		Ready = j->next;
		Ready = j->next;
		if (Ready == 0)
			ReadyEnd = 0;
		if (j->args) {
			exec_job(j);
			if (j->pid == 0) {
				warn("### skipping job: %s\n", j->cmd ?  j->cmd : j->args[0]);
			}
			free(j->out);
			free(j->err);
			free(j->in);
		} else {
			dbug("starting jobs after sequence point\n");
			assert(j != JobCache);
			j->next = JobCache;
			JobCache = j;
		}
	}
}


static void add_job(arg_t *a, size_t n, char *in, char *out, char *err, int app)
{
	char **argv, *exe;
	int i = n + Numbaseargs + 1;
	job_t *j;

	assert(a);
	assert(n > 0);
	j = JobCache;
	if (j)
		JobCache = j->next;
	else
		j = Malloc(sizeof(job_t));
	argv = Malloc(sizeof(char *) * i);
	j->in = in;
	j->err = err;
	j->out = out;
	j->append = app;
	j->outfile = -1;
	j->args = argv;
	j->pid = 0;
	j->next = 0;
	j->cmd = 0;
	--i;
	argv[i] = 0;
	j->numargs = i;
	while (n) {
		arg_t *c = a;
		assert(a);
		a = c->prev;
		--n;
		--i;
		argv[i] = c->arg;
		c->prev = ArgCache;
		ArgCache = c;
	}
	assert(a == 0);
	a = Baseargs;
	while (i) {
		arg_t *c = a;
		assert(a);
		a = c->prev;
		--i;
		argv[i] = c->arg;
	}
	assert(a == 0);
	exe = complete_exe(argv[0]);
	if (0 == exe) {
		char **a;
		warn("cannot find executable %s\n",argv[0]);
		a = &argv[Numbaseargs];
		while(*a)
			free(a++);
		free(argv);
		assert(j != JobCache);
		j->next = JobCache;
		JobCache = j;
		return;
	}
	if ((argv[0] != exe) && (Numbaseargs == 0))
		free(argv[0]);
	argv[0] = exe;
	if (Verbose >= Debug) {
		while (*argv) {
			(void) fprintf(stderr,"argument %d: <%s>\n",argv - j->args, *argv);
			++argv;
		}
	}
	j->next = 0;
	if (ReadyEnd)
		ReadyEnd->next = j;
	else
		Ready = j;
	ReadyEnd = j;
	++Waiting;
	start_jobs();
	if (Ready)
		dbug("job queued\n");
}


void clear_job(pid_t pid, int ret)
{
	job_t *j = Executing, *lj = 0;
	char **args;
	struct stat st;
	unsigned long long now;

	while (j && (j->pid != pid)) {
		lj = j;
		j = j->next;
	}
	if (j == 0) {
		/* consider the following bourne shell-script:
		 * #!/bin/sh
		 * ls -1 | ./a.out
		 *********************
		 * In this case a.out might be exec'ed by the shell
		 * _after_ forking childs for the other parts of the
		 * pipe. In consequence, depending on implementation,
		 * a.out can get a return value != -1 from wait that is
		 * associated with a process that cannot be found in our
		 * job list.  Be aware that the above script explicitly
		 * refers to /bin/sh not to /bin/bash. I.e. this issue
		 * concerns the original bourne shell.
		 */
		return;
	}
	now = gettime();
	if (lj)
		lj->next = j->next;
	else
		Executing = j->next;
	st.st_size = 0;
	if (j->outfile != -1) {
		int ret;
		ret = fstat(j->outfile,&st);
		if (ret != 0)
			warn("could not stat output file: %s\n",strerror(errno));
	}
	if (st.st_size > Pagesize) {
		void *mem = mmap(0,st.st_size,PROT_READ,MAP_SHARED,j->outfile,0);
		if (mem != (void *)-1) {
			(void) printf("### output of pid %ld: %s\n",pid,j->cmd);
			(void) fflush(stdout);
			(void) write(STDOUT_FILENO,mem,st.st_size);
			(void) printf("### exitcode %d\n",ret);
			(void) munmap(mem,st.st_size);
		} else {
			warn("could not mmap output of child: %s\n",strerror(errno));
			(void) printf("### pid %6ld, exitcode %d (%s) took %.3g secs.\n",pid,ret,j->cmd,(now-j->start)/1000000.0);
		}
	} else if (st.st_size > 0) {
		int r, s0, s1;
#if __STDC_VERSION__ < 199901L
		char *buf;
		buf = alloca(st.st_size + strlen(j->cmd) + 64);
#else
		char buf[st.st_size + strlen(j->cmd) + 64];
#endif
		s0 = sprintf(buf,"### output of pid %ld: %s\n",pid,j->cmd);
		r = pread(j->outfile,buf+s0,st.st_size,0);
		assert(r == st.st_size);
		s1 = sprintf(buf+s0+r,"### exitcode %d\n",ret);
		(void) write(STDOUT_FILENO,buf,s0+s1+r);
	} else {
		(void) printf("### pid %6ld, exitcode %d (%s) took %.3g secs.\n",pid,ret,j->cmd,(now-j->start)/1000000.0);
	}
	if (j->outfile != -1)
		(void) close(j->outfile);
	args = &j->args[Numbaseargs];
	while (*args) {
		free(*args);
		++args;
	}
	free(j->cmd);
	free(j->args);
	assert(j != JobCache);
	j->next = JobCache;
	JobCache = j;
	--Running;
}


static void usage(void)
{
	(void) printf(
	"xjobs [options] [command [argument ...]]\n"
	"========================================\n"
	"valid options are:\n"
	"-h           : print this usage information and exit\n"
	"-j <num>     : maximum number of jobs to execute in parallel\n"
	"-s <file>    : script to execute (default: read from stdin)\n"
	"-l <num>     : combine <num> lines to a single job\n"
	"-n           : redirect stdout and stderr of childs to /dev/null\n"
	"-v <level>   : set verbosity to level\n"
	"               (0=silent,1=error,2=warning,3=info,4=debug)\n"
	"-p           : prompt user, whether job should be started\n"
	"-V           : print version and exit\n"
	);
	exit(0);
}


static void version(void)
{
	(void) printf(
	"xjobs version 20070915\n"
	"Copyright 2006, Thomas Maier-Komor\n"
	"License: GPLv2\n"
	);
	exit(0);
}


int main(int argc, char **argv)
{
	typedef enum { init, std, white, dquoted, squoted, desc, sesc } state_t;
	int i, append = 0;
       	unsigned long lines = 1, lcount = 0;
	char *jobin = 0, *jobout = 0, *joberr = 0;
	extern char *optarg;
	extern int optind, optopt;
	arg_t *args = 0;
	size_t numargs = 0, q_len = 0;
	token_t t;

	Pagesize = sysconf(_SC_PAGESIZE);
	Path = getenv("PATH");
	while ((i = getopt(argc,argv,"hj:l:npq:s:v:V")) != -1) {
		switch (i) {
		default:
			abort();
			break;
		case 'h':
			usage();
			break;
		case 'j': {
				double lf;
				char f;
				switch (sscanf(optarg,"%lg%c",&lf,&f)) {
					case 2:
					if (f != 'x')
						error("invalid suffix in argument for option -j\n");
					lf *= sysconf(_SC_NPROCESSORS_ONLN);
					/*FALLTHRU*/
				case 1:
					Limit = ceil(lf);
					if (Limit <= 0)
						error("invalid argument for option -j\n");
					dbug("maximum number of jobs set to %d\n",Limit);
					break;
				default:
					error("missing argument to option -j\n");
				}
			} break;
		case 'l':
			if (1 == sscanf(optarg,"%lu",&lines) && (lines > 0))
				dbug("combining %lu lines to a single command\n",lines);
			else
				error("error in argument to option -l\n");
			break;
		case 'n':
			Stdout = open("/dev/null",O_WRONLY);
			if (Stdout == -1)
				error("could not open /dev/null: %s\n",strerror(errno));
			else
				dbug("output redirected to /dev/null\n");
			Stderr = Stdout;
			break;
		case 'p':
			Prompt = 1;
			Tty = fopen("/dev/tty","r+");
			if (Tty == 0)
				error("Cannot open terminal for prompt mode: %s\n",strerror(errno));
			close_onexec(fileno(Tty));
			dbug("enabling prompt mode\n");
			break;
		case 'q':
			if (1 == sscanf(optarg,"%u",&q_len) && (q_len > 0))
				dbug("limiting queue length to %lu elements\n",q_len);
			else
				error("error in argument to option -q\n");
			break;
		case 's':
			{
				extern FILE *yyin;
				FILE *in = fopen(optarg,"r");
				if (in == 0)
					error("could not open input file %s: %s\n",optarg,strerror(errno));
				close_onexec(fileno(in));
				yyin = in;
				dbug("input file set to %s\n",optarg);
				break;
			}
		case 'v':
			if (1 == sscanf(optarg,"%d",&Verbose)) {
				if (Verbose < Silent || Verbose > Debug)
					error("value out of range for option -v\n");
				dbug("verbose set to %d\n",Verbose);
			} else
				error("missing argument for option -v\n");
			break;
		case 'V':
			version();
			break;
		case '?':
			error("unkonwn option -%c\n",optopt);
			break;
		case ':':
			error("option -%c requires an operand\n",optopt);
			break;
		}
	}
        dbug("PATH=\"%s\"\n", Path);
	if (Limit == 0) {
		Limit = sysconf(_SC_NPROCESSORS_ONLN);
		dbug("%d processors currently online (default job limit)\n",Limit);
	}

	while (optind < argc) {
		if (0 == strcmp("<",argv[optind])) {
			if (Stdin != -1)
				(void) close(Stdin);
			Stdin = open(argv[++optind],O_RDONLY);
			if (-1 == Stdin)
				error("could not open default stdin for jobs (%s): %s\n",argv[optind],strerror(errno));
		} else if (0 == strcmp(">",argv[optind])) {
			if (Stdout != -1)
				(void) close(Stdout);
			Stdout = open(argv[++optind],O_WRONLY|O_CREAT|O_EXCL,0666);
			if (-1 == Stdout)
				error("could not open stdout for jobs (%s): %s\n",argv[optind],strerror(errno));
		} else if (0 == strcmp(">>",argv[optind])) {
			if (Stdout != -1)
				(void) close(Stdout);
			Stdout = open(argv[++optind],O_WRONLY|O_APPEND);
			if (-1 == Stdout)
				error("could not append to stdout for jobs (%s): %s\n",argv[optind],strerror(errno));
		} else if (0 == strcmp(">&",argv[optind])) {
			if (Stdout != -1)
				(void) close(Stdout);
			Stderr = Stdout = open(argv[++optind],O_WRONLY|O_CREAT|O_EXCL,0666);
			if (-1 == Stdout)
				error("could not open stdout/stderr for jobs (%s): %s\n",argv[optind],strerror(errno));
		} else if (0 == strcmp(">>&",argv[optind])) {
			if (Stdout != -1)
				(void) close(Stdout);
			Stdout = open(argv[++optind],O_WRONLY|O_APPEND);
			if (-1 == Stdout)
				error("could not append to stdout/stderr for jobs (%s): %s\n",argv[optind],strerror(errno));
		} else {
			arg_t *a = Malloc(sizeof(arg_t));
			if (Numbaseargs == 0) {
				a->arg = complete_exe(argv[optind]);
				if (a->arg == 0)
					error("cannot find %s in PATH\n",argv[optind]);
			} else {
				a->arg = argv[optind];
			}
			a->prev = Baseargs;
			Baseargs = a;
			++Numbaseargs;
		}
		++optind;
	}
	if (Stdin == -1) {	/* no default input redirector */
		Stdin = open("/dev/null",O_RDONLY);
		if (Stdin == -1)
			error("could not open /dev/null: %s\n",strerror(errno));
	}

	close_onexec(Stdin);
	close_onexec(Stdout);
	close_onexec(Stderr);

	while ((t = yylex())) {
		extern const char *Buffer;
		extern size_t Fill;
		pid_t pid;
		int ret;

		switch (t) {
		default:
			abort();
			break;
		case SEQUPT:
			{
				job_t *j = malloc(sizeof(job_t));
				j->args = 0;
				if (ReadyEnd)
					ReadyEnd->next = j;
				else
					Ready = j;
				ReadyEnd = j;
				break;
			}
		case INPUT:
			t = yylex();
			if ((t != UQUOTED) && (t != DQUOTED) && (t != SQUOTED)) {
				warn("invalid argument for input redirector <\n");
				do t = yylex(); while (t && (t != EOL));
			} else {
				jobin = Strdupn(Buffer,Fill+1);
			}
			break;
		case APPEND:
			append |= 1;
			/*FALLTHROUGH*/
		case OUTPUT:
			t = yylex();
			if ((t != UQUOTED) && (t != DQUOTED) && (t != SQUOTED)) {
				warn("invalid argument for output redirector >\n");
				do t = yylex(); while (t && (t != EOL));
			} else {
				jobout = Strdupn(Buffer,Fill+1);
			}
			break;
		case APPERR:
			append = 3;
			/*FALLTHROUGH*/
		case ERRLOG:
			t = yylex();
			if ((t != UQUOTED) && (t != DQUOTED) && (t != SQUOTED)) {
				warn("invalid argument for log redirector >&\n");
				do t = yylex(); while (t && (t != EOL));
			} else {
				joberr = Strdupn(Buffer,Fill+1);
			}
			break;
		case UQUOTED:
		case DQUOTED:
		case SQUOTED:
			{
				arg_t *a = ArgCache;
				if (a)
					ArgCache = a->prev;
				else
					a = Malloc(sizeof(arg_t));
				a->arg = Strdupn(Buffer,Fill+1);
				a->prev = args;
				args = a;
				++numargs;
				break;
			}
		case EOL:
			if (++lcount < lines)
				break;
			if (numargs)
				add_job(args,numargs,jobin,jobout,joberr,append);
			args = 0;
			numargs = 0;
			jobin = 0;
			jobout = 0;
			joberr = 0;
			append = 0;
			lcount = 0;
			break;
		case ERR:
			info("ignoring erroneous line\n");
			break;
		}
		
		if (Running) {
			if (pid = waitpid(0, &ret, q_len == Running ? 0 : WNOHANG)) {
				if (pid != (pid_t)-1)
					clear_job(pid,ret);
			}
		}
		start_jobs();
	}
	if (numargs)
		add_job(args,numargs,jobin,jobout,joberr,append);

	while (Executing || Ready) {
		pid_t pid;
		int ret;
		start_jobs();
		pid = wait(&ret);
		clear_job(pid,ret);
	}
	return 0;
}

